"""Atom class - contains all atom information found in the PDB file"""
import string
from propka.lib import make_tidy_atom_label
from . import hybrid36


# Format strings that get used in multiple places (or are very complex)
PKA_FMT = "{:6.2f}"
INPUT_LINE_FMT = (
    "{type:6s}{r.numb:>5d} {atom_label} {r.res_name}{r.chain_id:>2s}"
    "{r.res_num:>4d}{r.x:>12.3f}{r.y:>8.3f}{r.z:>8.3f}{group:>6s}{pka:>6s} \n")
PDB_LINE_FMT1 = (
    "{type:6s}{r.numb:>5d} {atom_label} {r.res_name}{r.chain_id:>2s}"
    "{r.res_num:>4d}{r.x:>12.3f}{r.y:>8.3f}{r.z:>8.3f}{r.occ:>6s}"
    "{r.beta:>6s}\n")
MOL2_LINE_FMT = (
    "{id:<4d} {atom_label:4s} "
    "{r.x:>10.4f} {r.y:>10.4f} {r.z:>10.4f} "
    "{r.sybyl_type:>6s} {r.res_num:>6d} {r.res_name:>10s}     0.0000\n")
PDB_LINE_FMT2 = (
    "ATOM {numb:>6d} {atom_label} {res_name}{chain_id:>2s}{res_num:>4d}"
    "{x:>12.3f}{y:>8.3f}{z:>8.3f}{occ:>6.2f}{beta:>6.2f}\n")
STR_FMT = (
    "{r.numb:>5d}-{r.name:>4s} {r.res_num:>5d}-{r.res_name:>3s} "
    "({r.chain_id:1s}) [{r.x:>8.3f} {r.y:>8.3f} {r.z:>8.3f}] {r.element:s}")


class Atom:
    """Atom class - contains all atom information found in the PDB file"""

    def __init__(self, line=None):
        """Initialize Atom object.

        Args:
            line:  Line from a PDB file to set properties of atom.
        """
        self.occ = None
        self.numb = None
        self.res_name = None
        self.type = None
        self.chain_id = None
        self.beta = None
        self.icode = None
        self.res_num = None
        self.name = None
        self.element = None
        self.x = None
        self.y = None
        self.z = None
        self.group = None
        self.group_type = None
        self.group_label = None
        self.group_model_pka = None
        self.group_model_pka_set = None
        self.number_of_bonded_elements = {}
        self.cysteine_bridge = False
        self.bonded_atoms = []
        self.residue = None
        self.conformation_container = None
        self.molecular_container = None
        self.is_protonated = False
        self.steric_num_lone_pairs_set = False
        self.terminal = None
        self.charge = 0
        self.charge_set = False
        self.steric_number = 0
        self.number_of_lone_pairs = 0
        self.number_of_protons_to_add = 0
        self.num_pi_elec_2_3_bonds = 0
        self.num_pi_elec_conj_2_3_bonds = 0
        self.groups_extracted = 0
        self.set_properties(line)
        fmt = "{r.name:3s}{r.res_num:>4d}{r.chain_id:>2s}"
        self.residue_label = fmt.format(r=self)

        # ligand atom types
        self.sybyl_type = ''
        self.sybyl_assigned = False
        self.marvin_pka = False

    def set_properties(self, line):
        """Line from PDB file to set properties of atom.

        Args:
            line:  PDB file line
        """
        self.name = ''
        self.numb = 0
        self.x = 0.0
        self.y = 0.0
        self.z = 0.0
        self.res_num = 0
        self.res_name = ''
        self.chain_id = 'A'
        self.type = ''
        self.occ = '1.0'
        self.beta = '0.0'
        self.element = ''
        self.icode = ''

        if line:
            self.name = line[12:16].strip()
            self.numb = int(hybrid36.decode(line[6:11]))
            self.x = float(line[30:38].strip())
            self.y = float(line[38:46].strip())
            self.z = float(line[46:54].strip())
            self.res_num = int(line[22:26].strip())
            self.res_name = "{0:<3s}".format(line[17:20].strip())
            self.chain_id = line[21]
            # Set chain id to "_" if it is just white space.
            if not self.chain_id.strip():
                self.chain_id = '_'
            self.type = line[:6].strip().lower()

            # TODO - define nucleic acid residue names elsewhere
            if self.res_name in ['DA ', 'DC ', 'DG ', 'DT ']:
                self.type = 'hetatm'

            self.occ = line[55:60].strip()
            self.beta = line[60:66].strip()
            self.icode = line[26:27]

            # Set the element using the position of the name in the pdb file
            self.element = line[12:14].strip().strip(string.digits)
            if len(self.name) == 4:
                self.element = self.element[0]
            if len(self.element) == 2:
                self.element = '{0:1s}{1:1s}'.format(
                    self.element[0], self.element[1].lower())

    def set_group_type(self, type_):
        """Set group type of atom.

        Args:
            type_:  group type of atom
        """
        self.group_type = type_

    def count_bonded_elements(self, element):
        """Count number of bonded atoms with same element.

        Args:
            element:  element type for test.
        Returns:
            number of bonded atoms.
        """
        return len(self.get_bonded_elements(element))

    def get_bonded_elements(self, element):
        """Get bonded atoms with same element.

        Args:
            element:  element type for test.
        Returns:
            array of bonded atoms.
        """
        res = []
        for bond_atom in self.bonded_atoms:
            if bond_atom.element == element:
                res.append(bond_atom)
        return res

    def get_bonded_heavy_atoms(self):
        """Get the atoms bonded to this one that aren't hydrogen.

        Returns:
            list of atoms.
        """
        return [ba for ba in self.bonded_atoms if ba.element != 'H']

    def is_atom_within_bond_distance(self, other_atom, max_bonds, cur_bond):
        """Check if <other_atom> is found within <max_bonds> bonds of self.

        Args:
            other_atom:  atom to check
            max_bonds:  number of bonds to check for other atom bonding to self
        Returns:
            Boolean for atom bond distance
        """
        for ba in self.bonded_atoms:
            if ba == other_atom:
                return True
            if max_bonds > cur_bond:
                if ba.is_atom_within_bond_distance(other_atom, max_bonds,
                                                   cur_bond+1):
                    return True
        return False

    def set_property(self, numb=None, name=None, res_name=None, chain_id=None,
                     res_num=None, x=None, y=None, z=None, occ=None,
                     beta=None):
        """Set properties of the atom object.

        Args:
            numb:  Atom number
            name:  Atom name
            res_name:  residue name
            chain_id:  chain ID
            res_num:  residue number
            x:  atom x-coordinate
            y:  atom y-coordinate
            z:  atom z-coordinate
            occ:  atom occupancy
            beta:  atom temperature factor
        """
        if numb is not None:
            self.numb = numb
        if name is not None:
            self.name = name
        if res_name is not None:
            self.res_name = res_name
        if chain_id is not None:
            self.chain_id = chain_id
        if res_num is not None:
            self.res_num = res_num
        if x is not None:
            self.x = x
        if y is not None:
            self.y = y
        if z is not None:
            self.z = z
        if occ is not None:
            self.occ = occ
        if beta is not None:
            self.beta = beta

    def make_copy(self):
        """Make a copy of this atom.

        Returns:
            Another atom object copy of this one.
        """
        new_atom = Atom()
        new_atom.type = self.type
        new_atom.numb = self.numb
        new_atom.name = self.name
        new_atom.element = self.element
        new_atom.res_name = self.res_name
        new_atom.res_num = self.res_num
        new_atom.chain_id = self.chain_id
        new_atom.x = self.x
        new_atom.y = self.y
        new_atom.z = self.z
        new_atom.occ = self.occ
        new_atom.beta = self.beta
        new_atom.terminal = self.terminal
        new_atom.residue_label = self.residue_label
        new_atom.icode = self.icode
        return new_atom

    def make_input_line(self):
        """PDB line for this atom.

        TODO - Could be @property method/attribute
        TODO - figure out difference between make_pdb_line, make_input_line,
               and make_pdb_line2

        Returns:
            String with PDB-format line.
        """
        group = '-'
        model_pka = '-'
        if self.group:
            group = self.group.type
            if self.terminal == 'C-':
                group = 'C-' ## circumventing C-/COO parameter unification

            if self.group.titratable:
                model_pka = PKA_FMT.format(self.group.model_pka)
        str_ = INPUT_LINE_FMT.format(
            type=self.type.upper(), r=self,
            atom_label=make_tidy_atom_label(self.name, self.element),
            group=group, pka=model_pka)
        return str_

    def make_conect_line(self):
        """PDB line for bonding within this molecule.

        Returns:
            String with PDB line.
        """
        res = 'CONECT{0:5d}'.format(self.numb)

        bonded = []
        for atom in self.bonded_atoms:
            bonded.append(atom.numb)
        bonded.sort()

        for bond in bonded:
            res += '{0:5d}'.format(bond)
        res += '\n'
        return res

    def get_input_parameters(self):
        """Extract the input parameters stored in the occupancy and b-factor
        fields in input files"""
        # Set the group type
        if self.occ != '-':
            # make sure to set the terminal
            if self.occ in ['N+', 'C-']:
                self.terminal = self.occ
            # save the ligand group charge
            if self.occ == 'BLG':
                self.charge = +1
            elif self.occ == 'ALG':
                self.charge = -1
            # generic ions
            if self.occ in ['1P', '2P', '1N', '2N']:
                self.res_name = self.occ
                self.occ = 'Ion'
            # correct the group type
            self.occ = self.occ.replace('N+', 'Nterm')
            self.occ = self.occ.replace('C-', 'Cterm')
            self.occ = self.occ.replace('ION', 'Ion')
            self.occ = self.occ.replace('ALG', 'titratable_ligand')
            self.occ = self.occ.replace('BLG', 'titratable_ligand')
            self.occ = self.occ.replace('LG', 'non_titratable_ligand')
            self.group_label = "{0:s}_group".format(self.occ)
        # set the model pKa value
        if self.beta != '-':
            self.group_model_pka = float(self.beta)
            self.group_model_pka_set = True
        # set occ and beta to standard values
        self.occ = '1.00'
        self.beta = '0.00'

    def make_pdb_line(self):
        """Create PDB line.

        TODO - this could/should be a @property method/attribute
        TODO - figure out difference between make_pdb_line, make_input_line,
               and make_pdb_line2

        Returns:
            String with PDB line.
        """
        str_ = PDB_LINE_FMT1.format(
            type=self.type.upper(), r=self,
            atom_label=make_tidy_atom_label(self.name, self.element))
        return str_

    def make_mol2_line(self, id_):
        """Create MOL2 line.

        Format: 1      S1     3.6147     2.0531     1.4795     S.3     1       noname  -0.1785

        TODO - this could/should be a @property method/attribute

        Returns:
            String with MOL2 line.
        """
        str_ = MOL2_LINE_FMT.format(
            id=id_, r=self,
            atom_label=make_tidy_atom_label(self.name, self.element))
        return str_

    def make_pdb_line2(self, numb=None, name=None, res_name=None, chain_id=None,
                       res_num=None, x=None, y=None, z=None, occ=None,
                       beta=None):
        """Create a PDB line.

        TODO - this could/should be a @property method/attribute
        TODO - figure out difference between make_pdb_line, make_input_line,
               and make_pdb_line2

        Returns:
            String with PDB line.
        """
        if numb is None:
            numb = self.numb
        if name is None:
            name = self.name
        if res_name is None:
            res_name = self.res_name
        if chain_id is None:
            chain_id = self.chain_id
        if res_num is None:
            res_num = self.res_num
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if z is None:
            z = self.z
        if occ is None:
            occ = self.occ
        if beta is None:
            beta = self.beta
        str_ = PDB_LINE_FMT2.format(
            numb=numb, res_name=res_name, chain_id=chain_id, res_num=res_num,
            x=x, y=y, z=z, occ=occ, beta=beta,
            atom_label=make_tidy_atom_label(name, self.element)
        )
        return str_

    def get_tidy_label(self):
        """Returns a 'tidier' atom label for printing the new pdbfile

        TODO - this could/should be a @property method/attribute

        Returns:
            String with label"""
        return make_tidy_atom_label(self.name, self.element)

    def __str__(self):
        """Return an undefined-format string version of this atom."""
        return STR_FMT.format(r=self)

    def set_residue(self, residue):
        """ Makes a reference to the parent residue

        Args:
            residue:  the parent residue
        """
        if self.residue is None:
            self.residue = residue



"""Iterative functions for pKa calculations.

These appear to mostly involve determinants.
"""
from propka.determinant import Determinant
from propka.lib import info, debug


# TODO - these are undocumented constants
UNK_MIN_VALUE = 0.005


def add_to_determinant_list(group1, group2, distance, iterative_interactions,
                            version):
    """Add iterative determinantes to the list.

    [[R1, R2], [side-chain, coulomb], [A1, A2]], ...

    NOTE - sign is determined when the interaction is added to the iterative
           object!
    NOTE - distance < coulomb_cutoff here

    Args:
        group1:  first group in pair
        group2:  second group in pair
        distance:  distance between groups
        iterative_interactions:  interaction list to modify
        version:  version object
    """
    hbond_value = version.hydrogen_bond_interaction(group1, group2)
    coulomb_value = version.electrostatic_interaction(group1, group2, distance)
    # adding the interaction to 'iterative_interactions'
    if hbond_value or coulomb_value:
        pair = [group1, group2]
        values = [hbond_value, coulomb_value]
        while None in values:
            values[values.index(None)] = 0.0
        annihilation = [0., 0.]
        interaction = [pair, values, annihilation]
        iterative_interactions.append(interaction)


def add_iterative_acid_pair(object1, object2, interaction):
    """Add the Coulomb 'iterative' interaction (an acid pair).

    The higher pKa is raised  with QQ+HB
    The lower pKa is lowered with HB

    Args:
        object1:  first object in pair
        object2:  second object in pair
        interaction:  list with [values, annihilation]
    """
    values = interaction[1]
    annihilation = interaction[2]
    hbond_value = values[0]
    coulomb_value = values[1]
    diff = coulomb_value + 2*hbond_value
    comp1 = object1.pka_old + annihilation[0] + diff
    comp2 = object2.pka_old + annihilation[1] + diff
    annihilation[0] = 0.0
    annihilation[1] = 0.0
    if comp1 > comp2:
        # side-chain
        determinant = [object2, hbond_value]
        object1.determinants['sidechain'].append(determinant)
        determinant = [object1, -hbond_value]
        object2.determinants['sidechain'].append(determinant)
        # Coulomb
        determinant = [object2, coulomb_value]
        object1.determinants['coulomb'].append(determinant)
        annihilation[0] = -diff
    else:
        # side-chain
        determinant = [object1, hbond_value]
        object2.determinants['sidechain'].append(determinant)
        determinant = [object2, -hbond_value]
        object1.determinants['sidechain'].append(determinant)
        # Coulomb
        determinant = [object1, coulomb_value]
        object2.determinants['coulomb'].append(determinant)
        annihilation[1] = -diff


def add_iterative_base_pair(object1, object2, interaction):
    """Add the Coulomb 'iterative' interaction (a base pair).

    The lower pKa is lowered

    Args:
        object1:  first object in pair
        object2:  second object in pair
        interaction:  list with [values, annihilation]
    """
    values = interaction[1]
    annihilation = interaction[2]
    hbond_value = values[0]
    coulomb_value = values[1]
    diff = coulomb_value + 2*hbond_value
    diff = -diff
    comp1 = object1.pka_old + annihilation[0] + diff
    comp2 = object2.pka_old + annihilation[1] + diff
    annihilation[0] = 0.0
    annihilation[1] = 0.0
    if comp1 < comp2:
        # side-chain
        determinant = [object2, -hbond_value]
        object1.determinants['sidechain'].append(determinant)
        determinant = [object1, hbond_value]
        object2.determinants['sidechain'].append(determinant)
        # Coulomb
        determinant = [object2, -coulomb_value]
        object1.determinants['coulomb'].append(determinant)
        annihilation[0] = -diff
    else:
        # side-chain
        determinant = [object1, -hbond_value]
        object2.determinants['sidechain'].append(determinant)
        determinant = [object2, hbond_value]
        object1.determinants['sidechain'].append(determinant)
        # Coulomb
        determinant = [object1, -coulomb_value]
        object2.determinants['coulomb'].append(determinant)
        annihilation[1] = -diff


def add_iterative_ion_pair(object1, object2, interaction, version):
    """Add the Coulomb 'iterative' interaction (an acid-base pair)

    the pKa of the acid is lowered & the pKa of the base is raised

    Args:
        object1:  first object in pair
        object2:  second object in pair
        interaction:  list with [values, annihilation]
        version:  version object
    """
    values = interaction[1]
    annihilation = interaction[2]
    hbond_value = values[0]
    coulomb_value = values[1]
    q1 = object1.q
    q2 = object2.q
    comp1 = object1.pka_old + annihilation[0] + q1*coulomb_value
    comp2 = object2.pka_old + annihilation[1] + q2*coulomb_value
    if (object1.res_name
            not in version.parameters.exclude_sidechain_interactions):
        comp1 += q1*hbond_value
    if (object2.res_name
            not in version.parameters.exclude_sidechain_interactions):
        comp2 += q2*hbond_value
    if q1 == -1.0 and comp1 < comp2:
        # pKa(acid) < pKa(base)
        add_term = True
    elif q1 == 1.0 and comp1 > comp2:
        # pKa(base) > pKa(acid)
        add_term = True
    else:
        add_term = False
    annihilation[0] = 0.00
    annihilation[1] = 0.00
    if add_term:
        # Coulomb
        if coulomb_value > UNK_MIN_VALUE:
            # residue1
            interaction = [object2, q1*coulomb_value]
            annihilation[0] += -q1*coulomb_value
            object1.determinants['coulomb'].append(interaction)
            # residue2
            interaction = [object1, q2*coulomb_value]
            annihilation[1] += -q2*coulomb_value
            object2.determinants['coulomb'].append(interaction)
        # Side-chain
        if hbond_value > UNK_MIN_VALUE:
            # residue1
            if (object1.res_name
                    not in version.parameters.exclude_sidechain_interactions):
                interaction = [object2, q1*hbond_value]
                annihilation[0] += -q1*hbond_value
                object1.determinants['sidechain'].append(interaction)
            # residue2
            if (object2.res_name
                    not in version.parameters.exclude_sidechain_interactions):
                interaction = [object1, q2*hbond_value]
                annihilation[1] += -q2*hbond_value
                object2.determinants['sidechain'].append(interaction)


def add_determinants(iterative_interactions, version, _=None):
    """Add determinants iteratively.

    The iterative pKa scheme. Later it is all added in 'calculateTotalPKA'

    Args:
        iterative_interactions:  list of iterative interactions
        version:  version object
        _:  options object
    """
    # --- setup ---
    iteratives = []
    done_group = []
    # create iterative objects with references to their real group counterparts
    for interaction in iterative_interactions:
        pair = interaction[0]
        for group in pair:
            if group in done_group:
                # do nothing - already have an iterative object for this group
                pass
            else:
                new_iterative = Iterative(group)
                iteratives.append(new_iterative)
                done_group.append(group)
    # Initialize iterative scheme
    debug(
        "\n   --- pKa iterations ({0:d} groups, {1:d} interactions) ---".format(
            len(iteratives), len(iterative_interactions)))
    converged = False
    iteration = 0
    # set non-iterative pka values as first step
    for iter_ in iteratives:
        iter_.pka_iter.append(iter_.pka_noniterative)
    # --- starting pKa iterations ---
    while not converged:
        # initialize pka_new
        iteration += 1
        for itres in iteratives:
            itres.determinants = {'sidechain': [], 'backbone': [],
                                  'coulomb': []}
            itres.pka_new = itres.pka_noniterative
        # Adding interactions to temporary determinant container
        for interaction in iterative_interactions:
            pair = interaction[0]
            object1, object2 = find_iterative(pair, iteratives)
            q1 = object1.q
            q2 = object2.q
            if   q1 < 0.0 and q2 < 0.0:
                # both are acids
                add_iterative_acid_pair(object1, object2, interaction)
            elif q1 > 0.0 and q2 > 0.0:
                # both are bases
                add_iterative_base_pair(object1, object2, interaction)
            else:
                # one of each
                add_iterative_ion_pair(object1, object2, interaction, version)
        # Calculating pka_new values
        for itres in iteratives:
            for type_ in ['sidechain', 'backbone', 'coulomb']:
                for determinant in itres.determinants[type_]:
                    itres.pka_new += determinant[1]

        # Check convergence
        converged = True
        for itres in iteratives:
            if itres.pka_new == itres.pka_old:
                itres.converged = True
            else:
                itres.converged = False
                converged = False

        # reset pka_old & storing pka_new in pka_iter
        for itres in iteratives:
            itres.pka_old = itres.pka_new
            itres.pka_iter.append(itres.pka_new)

        if iteration == 10:
            info("did not converge in {0:d} iterations".format(iteration))
            break
    # printing pKa iterations
    # formerly was conditioned on if options.verbosity >= 2 - now unnecessary
    str_ = '            '
    for index in range(iteration+1):
        str_ += "{0:>8d}".format(index)
    debug(str_)
    for itres in iteratives:
        str_ = "{0:s}   ".format(itres.label)
        for pka in itres.pka_iter:
            str_ += "{0:>8.2f}".format(pka)
        if not itres.converged:
            str_ += " *"
        debug(str_)
    # creating real determinants and adding them to group object
    for itres in iteratives:
        for type_ in ['sidechain', 'backbone', 'coulomb']:
            for interaction in itres.determinants[type_]:
                #info('done',itres.group.label,interaction[0],interaction[1])
                value = interaction[1]
                if value > UNK_MIN_VALUE or value < -UNK_MIN_VALUE:
                    group = interaction[0]
                    new_det = Determinant(group, value)
                    itres.group.determinants[type_].append(new_det)


def find_iterative(pair, iteratives):
    """Find the 'iteratives' that correspond to the groups in 'pair'.

    Args:
        pair:  groups to match
        iteratives:  list of iteratives to search
    Returns:
        1. first matched iterative
        2. second matched iterative
    """
    for iterative in iteratives:
        if iterative.group == pair[0]:
            iterative0 = iterative
        elif iterative.group == pair[1]:
            iterative1 = iterative
    return iterative0, iterative1


class Iterative:
    """Iterative class - pKa values and references of iterative groups.

    NOTE - this class has a fake determinant list, true determinants are made
    after the iterations are finished.
    """

    def __init__(self, group):
        """Initialize object with group.

        Args:
            group:  group to use for initialization.
        """
        self.label = group.label
        self.atom = group.atom
        self.res_name = group.residue_type
        self.q = group.charge
        self.pka_old = None
        self.pka_new = None
        self.pka_iter = []
        self.pka_noniterative = 0.00
        self.determinants = {'sidechain': [], 'backbone': [], 'coulomb': []}
        self.group = group
        self.converged = True
        # Calculate the Non-Iterative part of pKa from the group object
        # Side chain
        side_chain = 0.00
        for determinant in group.determinants['sidechain']:
            value = determinant.value
            side_chain += value
        # Back bone
        back_bone = 0.00
        for determinant in group.determinants['backbone']:
            value = determinant.value
            back_bone += value
        # Coulomb
        coulomb = 0.00
        for determinant in group.determinants['coulomb']:
            value = determinant.value
            coulomb += value
        self.pka_noniterative = group.model_pka
        self.pka_noniterative += group.energy_volume
        self.pka_noniterative += group.energy_local
        self.pka_noniterative += side_chain
        self.pka_noniterative += back_bone
        self.pka_noniterative += coulomb
        self.pka_old = self.pka_noniterative

    def __eq__(self, other):
        """Needed to use objects in sets."""
        if self.atom.type == 'atom':
            # In case of protein atoms we trust the labels
            return self.label == other.label
        else:
            # For heterogene atoms we also need to check the residue number
            return (
                self.label == other.label
                and self.atom.res_num == other.atom.res_num)

    def __hash__(self):
        """Needed to use objects in sets."""
        return id(self)



"""Container for molecular conformations"""
import functools
import propka.ligand
from propka.output import make_interaction_map
from propka.determinant import Determinant
from propka.coupled_groups import NCCG
from propka.determinants import set_backbone_determinants, set_ion_determinants
from propka.determinants import set_determinants
from propka.group import Group, is_group
from propka.lib import info


# A large number that gets multipled with the integer obtained from applying
# ord() to the atom chain ID.  Used in calculating atom keys for sorting.
UNICODE_MULTIPLIER = 1e7
# A number that gets mutiplied with an atom's residue number.  Used in
# calculating keys for atom sorting.
RESIDUE_MULTIPLIER = 1000


class ConformationContainer:
    """Container for molecular conformations"""

    def __init__(self, name='', parameters=None, molecular_container=None):
        """Initialize conformation container.

        Args:
            name:  name for conformation
            parameters:  parmameters for conformation
            molecular_container:  container for molecule
        """
        self.molecular_container = molecular_container
        self.name = name
        self.parameters = parameters
        self.atoms = []
        self.groups = []
        self.chains = []
        self.current_iter_item = 0
        self.marvin_pkas_calculated = False
        self.non_covalently_coupled_groups = False

    def extract_groups(self):
        """Generate molecular groups needed for calculating pKa values."""
        for atom in self.get_non_hydrogen_atoms():
            # has this atom been checked for groups?
            if atom.groups_extracted == 0:
                group = is_group(self.parameters, atom)
            else:
                group = atom.group
                # if the atom has been checked in a another conformation, check
                # if it has a group that should be used in this conformation as well
            if group:
                self.setup_and_add_group(group)

    def additional_setup_when_reading_input_file(self):
        """Generate interaction map and charge centers."""
        # if a group is coupled and we are reading a .propka_input file, then
        # some more configuration might be needed
        map_ = make_interaction_map(
            'Covalent coupling map for {0:s}'.format(str(self)),
            self.get_covalently_coupled_groups(),
            lambda g1, g2: g1 in g2.covalently_coupled_groups)
        info(map_)
        # check if we should set a common charge centre as well
        if self.parameters.common_charge_centre:
            self.set_common_charge_centres()

    def set_common_charge_centres(self):
        """Assign charge centers to groups."""
        for system in self.get_coupled_systems(
                self.get_covalently_coupled_groups(),
                Group.get_covalently_coupled_groups):
            # make a list of the charge centre coordinates
            all_coordinates = list(map(lambda g: [g.x, g.y, g.z], system))
            # find the common charge center
            ccc = functools.reduce(
                lambda g1, g2: [g1[0]+g2[0], g1[1]+g2[1], g1[2]+g2[2]],
                all_coordinates)
            ccc = list(map(lambda c: c/len(system), ccc))
            # set the ccc for all coupled groups in this system
            for group in system:
                [group.x, group.y, group.z] = ccc
                group.common_charge_centre = True

    def find_covalently_coupled_groups(self):
        """Find covalently coupled groups and set common charge centres."""
        for group in self.get_titratable_groups():
            # Find covalently bonded groups
            bonded_groups = self.find_bonded_titratable_groups(
                group.atom, 1, group.atom)
            # coupled groups
            for bond_group in bonded_groups:
                if bond_group in group.covalently_coupled_groups:
                    continue
                if bond_group.atom.sybyl_type == group.atom.sybyl_type:
                    group.couple_covalently(bond_group)
        # check if we should set a common charge centre as well
        if self.parameters.common_charge_centre:
            self.set_common_charge_centres()
        # print coupling map
        map_ = make_interaction_map(
            'Covalent coupling map for {0:s}'.format(str(self)),
            self.get_covalently_coupled_groups(),
            lambda g1, g2: g1 in g2.covalently_coupled_groups)
        info(map_)

    def find_non_covalently_coupled_groups(self, verbose=False):
        """Find non-covalently coupled groups and set common charge centres.

        Args:
            verbose:  verbose output
        """
        # check if non-covalent coupling has already been set up in an input file
        if len(list(filter(lambda g: len(g.non_covalently_coupled_groups) > 0,
                           self.get_titratable_groups()))) > 0:
            self.non_covalently_coupled_groups = True
        NCCG.identify_non_covalently_coupled_groups(self, verbose=verbose)
        # re-do the check
        if len(list(filter(lambda g: len(g.non_covalently_coupled_groups) > 0,
                           self.get_titratable_groups()))) > 0:
            self.non_covalently_coupled_groups = True

    def find_bonded_titratable_groups(self, atom, num_bonds, original_atom):
        """Find bonded titrable groups.

        Args:
            atom:  atom to check for bonds
            num_bonds:  number of bonds for coupling
            original_atom:  another atom to check for bonds
        Returns:
            a set of bonded atom groups
        """
        res = set()
        for bond_atom in atom.bonded_atoms:
            # skip the original atom
            if bond_atom == original_atom:
                continue
            # check if this atom has a titratable group
            if (bond_atom.group and bond_atom.group.titratable
                    and num_bonds
                    <= self.parameters.coupling_max_number_of_bonds):
                res.add(bond_atom.group)
            # check for titratable groups bonded to this atom
            if num_bonds < self.parameters.coupling_max_number_of_bonds:
                res |= self.find_bonded_titratable_groups(
                    bond_atom, num_bonds+1, original_atom)
        return res

    def setup_and_add_group(self, group):
        """Check if we want to include this group in the calculations.

        Args:
            group:  group to check
        """
        # Is it recognized as a group at all?
        if not group:
            return
        # Other checks (include ligands, which chains etc.)
        # if all ok, accept the group
        self.init_group(group)
        self.groups.append(group)

    def init_group(self, group):
        """Initialize the given Group object.

        Args:
            group:  group object to initialize
        """
        # set up the group
        group.parameters = self.parameters
        group.setup()

        # If --titrate_only option is set, make non-specified residues
        # un-titratable:
        titrate_only = self.molecular_container.options.titrate_only
        if titrate_only is not None:
            atom = group.atom
            if not (atom.chain_id, atom.res_num, atom.icode) in titrate_only:
                group.titratable = False
                if group.residue_type == 'CYS':
                    group.exclude_cys_from_results = True

    def calculate_pka(self, version, options):
        """Calculate pKas for conformation container.

        Args:
            version:  version object
            options:  option object
        """
        info('\nCalculating pKas for', self)
        # calculate desolvation
        for group in self.get_titratable_groups() + self.get_ions():
            version.calculate_desolvation(group)
        # calculate backbone interactions
        set_backbone_determinants(
            self.get_titratable_groups(), self.get_backbone_groups(), version)
        # setting ion determinants
        set_ion_determinants(self, version)
        # calculating the back-bone reorganization/desolvation term
        version.calculate_backbone_reorganization(self)
        # setting remaining non-iterative and iterative side-chain & Coulomb
        # interaction determinants
        set_determinants(
            self.get_sidechain_groups(), version=version, options=options)
        # calculating the total pKa values
        for group in self.groups:
            group.calculate_total_pka()
        # take coupling effects into account
        penalised_labels = self.coupling_effects()
        if (self.parameters.remove_penalised_group
                and len(penalised_labels) > 0):
            info('Removing penalised groups!!!')
            for group in self.get_titratable_groups():
                group.remove_determinants(penalised_labels)
            # re-calculating the total pKa values
            for group in self.groups:
                group.calculate_total_pka()

    def coupling_effects(self):
        """Penalize groups based on coupling effects.

        Bases: The group with the highest pKa (the most stable one in the
        charged form) will be the first one to adopt a proton as pH is lowered
        and this group is allowed to titrate. The remaining groups are
        penalised.

        Acids: The group with the highest pKa (the least stable one in the
        charged form) will be the last group to loose the proton as pH is
        raised and will be penalised. The remaining groups are allowed to
        titrate.
        """
        penalised_labels = []
        for all_groups in self.get_coupled_systems(
                self.get_covalently_coupled_groups(),
                Group.get_covalently_coupled_groups):
            # check if we should share determinants
            if self.parameters.shared_determinants:
                self.share_determinants(all_groups)
            # find the group that has the highest pKa value
            first_group = max(all_groups, key=lambda g: g.pka_value)
            # In case of acids
            if first_group.charge < 0:
                first_group.coupled_titrating_group = min(
                    all_groups, key=lambda g: g.pka_value)
                # group with the highest pKa is penalised
                penalised_labels.append(first_group.label)
            # In case of bases
            else:
                for group in all_groups:
                    if group == first_group:
                        # group with the highest pKa is allowed to titrate...
                        continue
                    group.coupled_titrating_group = first_group
                    #... and the rest are penalised
                    penalised_labels.append(group.label)
        return penalised_labels

    @staticmethod
    def share_determinants(groups):
        """Share sidechain, backbone, and Coloumb determinants between groups.

        Args:
            groups:  groups to share between
        """
        # make a list of the determinants to share
        types = ['sidechain', 'backbone', 'coulomb']
        for type_ in types:
            # find maximum value for each determinant
            max_dets = {}
            for group in groups:
                for det in group.determinants[type_]:
                    # update max dets
                    if det.group not in max_dets.keys():
                        max_dets[det.group] = det.value
                    else:
                        max_dets[det.group] = max(det.value,
                                                  max_dets[det.group],
                                                  key=lambda v: abs(v))
             # overwrite/add maximum value for each determinant
            for det_group in max_dets:
                new_determinant = Determinant(det_group, max_dets[det_group])
                for group in groups:
                    group.set_determinant(new_determinant, type_)

    def get_coupled_systems(self, groups, get_coupled_groups):
        """A generator that yields covalently coupled systems.

        Args:
            groups:  groups for generating coupled systems
            get_coupled_groups:  TODO - I don't know what this is
        Yields:
            covalently coupled systems
        """
        groups = set(groups)
        while len(groups) > 0:
            # extract a system of coupled groups ...
            system = set()
            self.get_a_coupled_system_of_groups(
                groups.pop(), system, get_coupled_groups)
            # ... and remove them from the list
            groups -= system
            yield system

    def get_a_coupled_system_of_groups(self, new_group, coupled_groups,
                                       get_coupled_groups):
        """Set up coupled systems of groups.

        Args:
            new_group:  added to coupled_groups
            coupled_groups:  existing coupled groups
            get_coupled_groups:  TODO - I don't know what this
        """
        coupled_groups.add(new_group)
        for coupled_group in get_coupled_groups(new_group):
            if coupled_group not in coupled_groups:
                self.get_a_coupled_system_of_groups(coupled_group,
                                                    coupled_groups,
                                                    get_coupled_groups)

    def calculate_folding_energy(self, ph=None, reference=None):
        """Calculate folding energy over all groups in conformation container.

        Args:
            ph:  pH for calculation
            reference:  reference state
        Returns:
            folding energy
            TODO - need units
        """
        ddg = 0.0
        for group in self.groups:
            ddg += group.calculate_folding_energy(self.parameters, ph=ph,
                                                  reference=reference)
        return ddg

    def calculate_charge(self, parameters, ph=None):
        """Calculate charge for folded and unfolded states.

        Args:
            parameters:  parameters for calculation
            ph:  pH for calculation
        Returns:
            1. charge for unfolded state
            2. charge for folded state
        """
        unfolded = folded = 0.0
        for group in self.get_titratable_groups():
            unfolded += group.calculate_charge(parameters, ph=ph,
                                               state='unfolded')
            folded += group.calculate_charge(parameters, ph=ph,
                                             state='folded')
        return unfolded, folded

    def get_backbone_groups(self):
        """Get backbone groups needed for the pKa calculations.

        Returns:
            list of groups
        """
        return [group for group in self.groups if 'BB' in group.type]

    def get_sidechain_groups(self):
        """Get sidechain groups needed for the pKa calculations.

        Returns:
            list of groups
        """
        return [
            group for group in self.groups
            if ('BB' not in group.type and not group.atom.cysteine_bridge)]

    def get_covalently_coupled_groups(self):
        """Get covalently coupled groups needed for pKa calculations.

        Returns:
            list of groups
        """
        return [
            g for g in self.groups
            if len(g.covalently_coupled_groups) > 0]

    def get_non_covalently_coupled_groups(self):
        """Get non-covalently coupled groups needed for pKa calculations.

        Returns:
            list of groups
        """
        return [
            g for g in self.groups
            if len(g.non_covalently_coupled_groups) > 0]

    def get_backbone_nh_groups(self):
        """Get NH backbone groups needed for pKa calculations.

        Returns:
            list of groups
        """
        return [group for group in self.groups if group.type == 'BBN']

    def get_backbone_co_groups(self):
        """Get CO backbone groups needed for pKa calculations.

        Returns:
            list of groups
        """
        return [group for group in self.groups if group.type == 'BBC']

    def get_groups_in_residue(self, residue):
        """Get residue groups needed for pKa calculations.

        Args:
            residue:  specific residue with groups
        Returns:
            list of groups
        """
        return [
            group for group in self.groups if group.residue_type == residue]

    def get_titratable_groups(self):
        """Get all titratable groups needed for pKa calculations.

        Returns:
            list of groups
        """
        return [group for group in self.groups if group.titratable]

    def get_groups_for_calculations(self):
        """Get a list of groups that should be included in results report.

        If --titrate_only option is specified, only residues that are titratable
        and are in that list are included; otherwise all titratable residues
        and CYS residues are included.

        Returns:
            list of groups
        """
        return [group for group in self.groups if group.use_in_calculations()]

    def get_acids(self):
        """Get acid groups needed for pKa calculations.

        Returns:
            list of groups
        """
        return [
            group for group in self.groups
            if (group.residue_type in self.parameters.acid_list
                and not group.atom.cysteine_bridge)]

    def get_backbone_reorganisation_groups(self):
        """Get groups involved with backbone reorganization.

        Returns:
            list of groups
        """
        return [
            group for group in self.groups
            if (group.residue_type
                in self.parameters.backbone_reorganisation_list
                and not group.atom.cysteine_bridge)]

    def get_ions(self):
        """Get ion groups.

        Returns:
            list of groups
        """
        return [
            group for group in self.groups
            if group.residue_type in self.parameters.ions.keys()]

    def get_group_names(self, group_list):
        """Get names of groups in list.

        Args:
            group_list:  list to check
        Returns:
            list of groups
        """
        return [group for group in self.groups if group.type in group_list]

    def get_ligand_atoms(self):
        """Get atoms associated with ligands.

        Returns:
            list of atoms
        """
        return [atom for atom in self.atoms if atom.type == 'hetatm']

    def get_heavy_ligand_atoms(self):
        """Get heavy atoms associated with ligands.

        Returns:
            list of atoms
        """
        return [
            atom for atom in self.atoms
            if atom.type == 'hetatm' and atom.element != 'H']

    def get_chain(self, chain):
        """Get atoms associated with a specific chain.

        Args:
            chain:  chain to select
        Returns:
            list of atoms
        """
        return [atom for atom in self.atoms if atom.chain_id != chain]

    def add_atom(self, atom):
        """Add atom to container.

        Args:
            atom:  atom to add
        """
        self.atoms.append(atom)
        if not atom.conformation_container:
            atom.conformation_container = self
        if not atom.molecular_container:
            atom.molecular_container = self.molecular_container
        # store chain id for bookkeeping
        if not atom.chain_id in self.chains:
            self.chains.append(atom.chain_id)

    def copy_atom(self, atom):
        """Add a copy of the atom to container.

        Args:
            atom:  atom to copy and add
        """
        new_atom = atom.make_copy()
        self.atoms.append(new_atom)
        new_atom.conformation_container = self

    def get_non_hydrogen_atoms(self):
        """Get atoms that are not hydrogens.

        Returns:
            list of atoms
        """
        return [atom for atom in self.atoms if atom.element != 'H']

    def top_up(self, other):
        """Adds any atoms found in `other` but not in this container.

        Tops up self with all atoms found in other but not in self.

        Args:
            other:  conformation container with atoms to add
        """
        my_residue_labels = {a.residue_label for a in self.atoms}
        for atom in other.atoms:
            if not atom.residue_label in my_residue_labels:
                self.copy_atom(atom)

    def find_group(self, group):
        """Find a group in the container.

        Args:
            group:  group to find
        Returns:
            False (if group not found) or group
        """
        for group_ in self.groups:
            if group_.atom.residue_label == group.atom.residue_label:
                if group_.type == group.type:
                    return group_
        return False

    def set_ligand_atom_names(self):
        """Set names for atoms in ligands."""
        for atom in self.get_ligand_atoms():
            propka.ligand.assign_sybyl_type(atom)

    def __str__(self):
        """String that lists statistics of atoms and groups."""
        fmt = (
            "Conformation container {name} with {natoms:d} atoms and "
            "{ngroups:d} groups")
        str_ = fmt.format(
            name=self.name, natoms=len(self), ngroups=len(self.groups))
        return str_

    def __len__(self):
        """Number of atoms in container."""
        return len(self.atoms)

    def sort_atoms(self):
        """Sort atoms by `self.sort_atoms_key()` and renumber."""
        # sort the atoms ...
        self.atoms.sort(key=self.sort_atoms_key)
        # ... and re-number them
        for i in range(len(self.atoms)):
            self.atoms[i].numb = i+1

    @staticmethod
    def sort_atoms_key(atom):
        """Generate key for atom sorting.

        Args:
            atom:  atom for key generation.
        Returns:
            key for atom
        """
        key = ord(atom.chain_id) * UNICODE_MULTIPLIER
        key += atom.res_num * RESIDUE_MULTIPLIER
        if len(atom.name) > len(atom.element):
            key += ord(atom.name[len(atom.element)])
        return key



"""Molecular container for storing all contents of PDB files."""
import os
import propka.version
from propka.output import write_propka, write_pka, print_header, print_result
from propka.conformation_container import ConformationContainer
from propka.lib import info, warning, make_grid


# TODO - these are constants whose origins are a little murky
UNK_PI_CUTOFF = 0.01
# Maximum number of iterations for finding PI
MAX_ITERATION = 4


class MolecularContainer:
    """Container for storing molecular contents of PDB files.

    TODO - this class name does not conform to PEP8 but has external use.
    We should deprecate and change eventually.
    """

    def __init__(self, parameters, options=None):
        """Initialize molecular container.

        Args:
            parameters:  Parameters() object
            options:  options object
        """
        # printing out header before parsing input
        print_header()
        self.conformation_names = []
        self.conformations = {}
        self.options = options
        self.name = None
        try:
            version_class = getattr(propka.version, parameters.version)
            self.version = version_class(parameters)
        except AttributeError as err:
            print(err)
            errstr = 'Error: Version {0:s} does not exist'.format(
                parameters.version)
            raise Exception(errstr)

    def top_up_conformations(self):
        """Makes sure that all atoms are present in all conformations."""
        for name in self.conformation_names:
            if (name != '1A' and (len(self.conformations[name])
                                  < len(self.conformations['1A']))):
                self.conformations[name].top_up(self.conformations['1A'])

    def find_covalently_coupled_groups(self):
        """Find covalently coupled groups."""
        info('-' * 103)
        for name in self.conformation_names:
            self.conformations[name].find_covalently_coupled_groups()

    def find_non_covalently_coupled_groups(self):
        """Find non-covalently coupled groups."""
        info('-' * 103)
        verbose = self.options.display_coupled_residues
        for name in self.conformation_names:
            self.conformations[name].find_non_covalently_coupled_groups(
                verbose=verbose)

    def extract_groups(self):
        """Identify the groups needed for pKa calculation."""
        for name in self.conformation_names:
            self.conformations[name].extract_groups()

    def additional_setup_when_reading_input_file(self):
        """Additional setup."""
        for name in self.conformation_names:
            self.conformations[name].additional_setup_when_reading_input_file()

    def calculate_pka(self):
        """Calculate pKa values."""
        # calculate for each conformation
        for name in self.conformation_names:
            self.conformations[name].calculate_pka(
                self.version, self.options)
        # find non-covalently coupled groups
        self.find_non_covalently_coupled_groups()
        # find the average of the conformations
        self.average_of_conformations()
        # print out the conformation-average results
        print_result(self, 'AVR', self.version.parameters)

    def average_of_conformations(self):
        """Generate an average of conformations."""
        parameters = self.conformations[self.conformation_names[0]].parameters
        # make a new configuration to hold the average values
        avr_conformation = ConformationContainer(
            name='average', parameters=parameters, molecular_container=self)
        container = self.conformations[self.conformation_names[0]]
        for group in container.get_groups_for_calculations():
            # new group to hold average values
            avr_group = group.clone()
            # sum up all groups ...
            for name in self.conformation_names:
                group_to_add = self.conformations[name].find_group(group)
                if group_to_add:
                    avr_group += group_to_add
                else:
                    str_ = (
                        'Group {0:s} could not be found in '
                        'conformation {1:s}.'.format(
                            group.atom.residue_label, name))
                    warning(str_)
            # ... and store the average value
            avr_group = avr_group / len(self.conformation_names)
            avr_conformation.groups.append(avr_group)
        # store information on coupling in the average container
        if len(list(filter(lambda c: c.non_covalently_coupled_groups,
                           self.conformations.values()))):
            avr_conformation.non_covalently_coupled_groups = True
        # store chain info
        avr_conformation.chains = self.conformations[
            self.conformation_names[0]].chains
        self.conformations['AVR'] = avr_conformation

    def write_propka(self, filename=None):
        """Write PROPKA input file.

        Args:
            filename:  file to write to
        """
        if filename is None:
            filename = os.path.join('{0:s}.propka_input'.format(self.name))
        write_propka(self, filename)

    def write_pka(self, filename=None, reference="neutral",
                  direction="folding", options=None):
        """Write pKa information to a file.

        Args:
            filename:  file to write to
            reference:  reference state
            direction:  folding vs. unfolding
            options:  options object
        """
        if filename is None:
            filename = os.path.join('{0:s}.pka'.format(self.name))
        # if the display_coupled_residues option is true, write the results out
        # to an alternative pka file
        if self.options.display_coupled_residues:
            filename = os.path.join('{0:s}_alt_state.pka'.format(self.name))
        if (hasattr(self.version.parameters, 'output_file_tag')
                and len(self.version.parameters.output_file_tag) > 0):
            filename = os.path.join(
                '{0:s}_{1:s}.pka'.format(
                    self.name, self.version.parameters.output_file_tag))
        write_pka(
            self, self.version.parameters, filename=filename,
            conformation='AVR', reference=reference)

    def get_folding_profile(self, conformation='AVR', reference="neutral",
                            grid=[0., 14., 0.1]):
        """Get a folding profile.

        Args:
            conformation:  conformation to select
            reference:  reference state
            direction:  folding direction (folding)
            grid:  the grid of pH values [min, max, step_size]
            options:  options object
        Returns:
            TODO - figure out what these are
            1. profile
            2. opt
            3. range_80pct
            4. stability_range
        """
        # calculate stability profile
        profile = []
        for ph in make_grid(*grid):
            conf = self.conformations[conformation]
            ddg = conf.calculate_folding_energy(ph=ph, reference=reference)
            profile.append([ph, ddg])
        # find optimum
        opt = [None, 1e6]
        for point in profile:
            opt = min(opt, point, key=lambda v: v[1])
        # find values within 80 % of optimum
        range_80pct = [None, None]
        values_within_80pct = [p[0] for p in profile if p[1] < 0.8*opt[1]]
        if len(values_within_80pct) > 0:
            range_80pct = [min(values_within_80pct), max(values_within_80pct)]
        # find stability range
        stability_range = [None, None]
        stable_values = [p[0] for p in profile if p[1] < 0.0]
        if len(stable_values) > 0:
            stability_range = [min(stable_values), max(stable_values)]
        return profile, opt, range_80pct, stability_range

    def get_charge_profile(self, conformation='AVR', grid=[0., 14., .1]):
        """Get charge profile for conformation as function of pH.

        Args:
            conformation:  conformation to test
            grid:  grid of pH values [min, max, step]
        Returns:
            list of charge state values
        """
        charge_profile = []
        for ph in make_grid(*grid):
            conf = self.conformations[conformation]
            q_unfolded, q_folded = conf.calculate_charge(
                self.version.parameters, ph=ph)
            charge_profile.append([ph, q_unfolded, q_folded])
        return charge_profile

    def get_pi(self, conformation='AVR', grid=[0., 14., 1], iteration=0):
        """Get the isoelectric points for folded and unfolded states.

        Args:
            conformation:  conformation to test
            grid:  grid of pH values [min, max, step]
            iteration:  iteration number of process
        Returns:
            1. Folded state PI
            2. Unfolded state PI
        """
        charge_profile = self.get_charge_profile(
            conformation=conformation, grid=grid)
        pi_folded = pi_unfolded = [None, 1e6, 1e6]
        for point in charge_profile:
            pi_folded = min(pi_folded, point, key=lambda v: abs(v[2]))
            pi_unfolded = min(pi_unfolded, point, key=lambda v: abs(v[1]))
        # If results are not good enough, do it again with a higher sampling
        # resolution
        pi_folded_value = pi_folded[0]
        pi_unfolded_value = pi_unfolded[0]
        step = grid[2]
        # TODO - need to warn if maximum number of iterations is exceeded
        if ((pi_folded[2] > UNK_PI_CUTOFF
             or pi_unfolded[1] > UNK_PI_CUTOFF) and iteration < MAX_ITERATION):
            pi_folded_value, _ = self.get_pi(
                conformation=conformation,
                grid=[pi_folded[0]-step, pi_folded[0]+step, step/10.0],
                iteration=iteration+1)
            _, pi_unfolded_value = self.get_pi(
                conformation=conformation,
                grid=[pi_unfolded[0]-step, pi_unfolded[0]+step, step/10.0],
                iteration=iteration+1)
        return pi_folded_value, pi_unfolded_value



"""Describe coupling between groups."""
import itertools
import propka.lib
from propka.group import Group
from propka.output import make_interaction_map
from propka.lib import info


class NonCovalentlyCoupledGroups:
    """Groups that are coupled without covalent bonding."""
    def __init__(self):
        self.parameters = None
        self.do_prot_stat = True

    def is_coupled_protonation_state_probability(self, group1, group2,
                                                 energy_method,
                                                 return_on_fail=True):
        """Check whether two groups are energetically coupled.

        Args:
            group1:  first group for interaction
            group2:  second group for interaction
            energy_method:  function for calculating energy
            return_on_fail:  return if part of the calculation fails
        Returns:
            dictionary describing coupling
        """
        # check if the interaction energy is high enough
        interaction_energy = max(self.get_interaction(group1, group2),
                                 self.get_interaction(group2, group1))
        if (interaction_energy <= self.parameters.min_interaction_energy
                and return_on_fail):
            return {'coupling_factor': -1.0}
         # calculate intrinsic pKa's, if not already done
        for group in [group1, group2]:
            if group.intrinsic_pka is None:
                group.calculate_intrinsic_pka()
        use_ph = self.parameters.pH
        if self.parameters.pH == 'variable':
            use_ph = min(group1.pka_value, group2.pka_value)
        default_energy = energy_method(ph=use_ph,
                                       reference=self.parameters.reference)
        default_pka1 = group1.pka_value
        default_pka2 = group2.pka_value
        # check that pka values are within relevant limits
        if (max(default_pka1, default_pka2) < self.parameters.min_pka
                or min(default_pka1, default_pka2) > self.parameters.max_pka):
            if return_on_fail:
                return {'coupling_factor': -1.0}
        # Swap interactions and re-calculate pKa values
        self.swap_interactions([group1], [group2])
        group1.calculate_total_pka()
        group2.calculate_total_pka()
        # store swapped energy and pka's
        swapped_energy = energy_method(
            ph=use_ph, reference=self.parameters.reference)
        swapped_pka1 = group1.pka_value
        swapped_pka2 = group2.pka_value
        pka_shift1 = swapped_pka1 - default_pka1
        pka_shift2 = swapped_pka2 - default_pka2
        # Swap back to original protonation state
        self.swap_interactions([group1], [group2])
        group1.calculate_total_pka()
        group2.calculate_total_pka()
        # check difference in free energy
        if (abs(default_energy - swapped_energy)
                > self.parameters.max_free_energy_diff and return_on_fail):
            return {'coupling_factor': -1.0}
        # check pka shift
        if (max(abs(pka_shift1), abs(pka_shift2))
                < self.parameters.min_swap_pka_shift and return_on_fail):
            return {'coupling_factor': -1.0}
        # check intrinsic pka diff
        if (abs(group1.intrinsic_pka - group2.intrinsic_pka)
                > self.parameters.max_intrinsic_pka_diff and return_on_fail):
            return {'coupling_factor': -1.0}
        # if everything is OK, calculate the coupling factor and return all info
        factor = (
            self.get_free_energy_diff_factor(default_energy, swapped_energy)
            * self.get_pka_diff_factor(group1.intrinsic_pka,
                                       group2.intrinsic_pka)
            * self.get_interaction_factor(interaction_energy))
        return {'coupling_factor': factor, 'default_energy': default_energy,
                'swapped_energy': swapped_energy,
                'interaction_energy': interaction_energy,
                'swapped_pka1': swapped_pka1, 'swapped_pka2': swapped_pka2,
                'pka_shift1': pka_shift1, 'pka_shift2': pka_shift2,
                'pH': use_ph}

    def get_pka_diff_factor(self, pka1, pka2):
        """Get scaling factor for difference between intrinsic pKa values.

        Args:
            pka1:  first pKa to compare
            pka2:  second pKa to compare
        Returns:
            float value of scaling factor
        """
        intrinsic_pka_diff = abs(pka1-pka2)
        res = 0.0
        if intrinsic_pka_diff <= self.parameters.max_intrinsic_pka_diff:
            res = (
                1-(intrinsic_pka_diff
                   /self.parameters.max_intrinsic_pka_diff)**2)
        return res

    def get_free_energy_diff_factor(self, energy1, energy2):
        """Get scaling factor for difference between free energies.

        Args:
            energy1:  first energy to compare
            energy2:  second energy to compare
        Returns:
            float value of scaling factor
        """
        free_energy_diff = abs(energy1-energy2)
        res = 0.0
        if free_energy_diff <= self.parameters.max_free_energy_diff:
            res = 1-(free_energy_diff/self.parameters.max_free_energy_diff)**2
        return res

    def get_interaction_factor(self, interaction_energy):
        """Get scaling factor related to interaction energy.

        Args:
            interaction_energy:  interaction energy
        Returns:
            float value of scaling factor
        """
        res = 0.0
        interaction_energy = abs(interaction_energy)
        if interaction_energy >= self.parameters.min_interaction_energy:
            res = (
                (interaction_energy-self.parameters.min_interaction_energy)
                / (1.0+interaction_energy
                   -self.parameters.min_interaction_energy))
        return res

    def identify_non_covalently_coupled_groups(self, conformation,
                                               verbose=True):
        """Find coupled residues in protein.

        Args:
            conformation:  protein conformation to test
            verbose:  verbose output (boolean)
        """
        self.parameters = conformation.parameters
        if verbose:
            info_fmt = (
                '\n'
                ' Warning: When using the -d option, pKa values based on \n'
                '\'swapped\' interactions\n'
                '          will be writting to the output .pka file\n'
                '\n'
                '{sep}\n'
                '\n'
                ' Detecting non-covalently coupled residues\n'
                '{sep}\n'
                '   Maximum pKa difference:     {c.max_intrinsic_pka_diff:>4.2f} pKa units\n'
                '   Minimum interaction energy: {c.min_interaction_energy:>4.2f} pKa units\n'
                '   Maximum free energy diff.:  {c.max_free_energy_diff:>4.2f} pKa units\n'
                '   Minimum swap pKa shift:     {c.min_swap_pka_shift:>4.2f} pKa units\n'
                '   pH:                         {c.pH:>6} \n'
                '   Reference:                  {c.reference}\n'
                '   Min pKa:                    {c.min_pka:>4.2f}\n'
                '   Max pKa:                    {c.max_pka:>4.2f}\n'
                '\n')
            sep = "-" * 103
            info(info_fmt.format(sep=sep, c=self))
        # find coupled residues
        titratable_groups = conformation.get_titratable_groups()
        if not conformation.non_covalently_coupled_groups:
            for group1 in titratable_groups:
                for group2 in titratable_groups:
                    if group1 == group2:
                        break
                    if (group1 not in group2.non_covalently_coupled_groups
                            and self.do_prot_stat):
                        data = (
                            self
                            .is_coupled_protonation_state_probability(
                                group1, group2,
                                conformation.calculate_folding_energy))
                        if data['coupling_factor'] > 0.0:
                            group1.couple_non_covalently(group2)
        if verbose:
            self.print_out_swaps(conformation)

    def print_out_swaps(self, conformation):
        """Print out something having to do with coupling interactions.

        Args:
            conformation:  conformation to print
        """
        map_ = make_interaction_map(
            'Non-covalent coupling map for {0:s}'.format(str(conformation)),
            conformation.get_non_covalently_coupled_groups(),
            lambda g1, g2: g1 in g2.non_covalently_coupled_groups)
        info(map_)
        for system in conformation.get_coupled_systems(
                conformation.get_non_covalently_coupled_groups(),
                Group.get_non_covalently_coupled_groups):
            self.print_system(conformation, list(system))

    def print_system(self, conformation, system):
        """Print out something about the system.

        Args:
            conformation:  conformation to print
            system:  system to print
        """
        info(
            'System containing {0:d} groups:'.format(len(system)))
        # make list of interactions within this system
        interactions = list(itertools.combinations(system, 2))
        # print out coupling info for each interaction
        coup_info = ''
        for interaction in interactions:
            data = (
                self.is_coupled_protonation_state_probability(
                    interaction[0], interaction[1],
                    conformation.calculate_folding_energy,
                    return_on_fail=False))
            coup_info += (
                self.make_data_to_string(data, interaction[0], interaction[1])
                + '\n\n')
        info(coup_info)
        # make list of possible combinations of swap to try out
        combinations = propka.lib.generate_combinations(interactions)
        # Make possible swap combinations
        swap_info = ''
        swap_info += self.print_determinants_section(system, 'Original')
        for combination in combinations:
            # Tell the user what is swap in this combination
            swap_info += 'Swapping the following interactions:\n'
            for interaction in combination:
                swap_info += ' {0:s} {1:s}\n'.format(
                    interaction[0].label, interaction[1].label)
            # swap...
            for interaction in combination:
                self.swap_interactions([interaction[0]], [interaction[1]])
            swap_info += self.print_determinants_section(system, 'Swapped')
        info(swap_info)

    @staticmethod
    def get_interaction(group1, group2, include_side_chain_hbs=True):
        """Get interaction energy between two groups.

        Args:
            group1:  first group for interaction
            group2:  second group for interaction
            include_side_chain_hbs:  include sidechain hydrogen bonds in energy
        Returns:
            interaction energy (float)
        """
        determinants = group1.determinants['coulomb']
        if include_side_chain_hbs:
            determinants = (
                group1.determinants['sidechain']
                + group1.determinants['coulomb'])
        interaction_energy = 0.0
        for det in determinants:
            if group2 == det.group:
                interaction_energy += det.value
        return interaction_energy

    def print_determinants_section(self, system, tag):
        """Print determinants of system.

        Args:
            system:  set of groups
            tag:  something to add to output
        Returns:
            string with summary
        """
        all_labels = [g.label for g in system]
        str_ = ' ' + '-' * 113 + '\n'
        for group in system:
            str_ += self.tagged_format(
                ' {0:<8s}|'.format(tag), group.get_determinant_string(),
                all_labels)
        return str_ + '\n'

    def swap_interactions(self, groups1, groups2, include_side_chain_hbs=True):
        """Swap interactions between two groups.

        Args:
            group1:  first group to swap
            group2:  second group to swap
            """
        for i, group1 in enumerate(groups1):
            group2 = groups2[i]
            # swap the interactions!
            self.transfer_determinant(group1.determinants['coulomb'],
                                      group2.determinants['coulomb'],
                                      group1.label, group2.label)
            if include_side_chain_hbs:
                self.transfer_determinant(group1.determinants['sidechain'],
                                          group2.determinants['sidechain'],
                                          group1.label, group2.label)
                # re-calculate pKa values
                group1.calculate_total_pka()
                group2.calculate_total_pka()

    @staticmethod
    def transfer_determinant(determinants1, determinants2,
                             label1, label2):
        """Transfer information between two sets of determinants.

        Args:
            determinants1:  determinant list
            determinants2:  determinant list
            label1:  label for list 1
            label2:  label for list 2
        """
        # find out what to transfer...
        from1to2 = []
        from2to1 = []
        for det in determinants1:
            if det.label == label2:
                from1to2.append(det)
        for det in determinants2:
            if det.label == label1:
                from2to1.append(det)
        # ...and transfer it!
        for det in from1to2:
            det.label = label1
            determinants2.append(det)
            determinants1.remove(det)
        for det in from2to1:
            det.label = label2
            determinants1.append(det)
            determinants2.remove(det)

    @staticmethod
    def tagged_format(tag, str_, labels):
        """Tag a string.

        Args:
            tag:  tag to add
            str_:  string to tag
            labels:  labels to replace
        Returns:
            tagged string
        """
        str_ = "{0:s} {1:s}".format(tag, str_)
        str_ = str_.replace('\n', '\n{0:s} '.format(tag))
        for label in labels:
            str_ = str_.replace(label, '\033[31m{0:s}\033[30m'.format(label))
        return str_ + '\n'

    @staticmethod
    def make_data_to_string(data, group1, group2):
        """Describe interaction between groups.

        Args:
            data:  data about interactions
            group1:  first group
            group2:  second group
        Returns:
            formatted string with information.
        """
        str_ = (
            " {label1} and {label2} coupled (prot.state): {coupl_fact:>5.2f}\n"
            " Energy levels:       {def_energy:>6.2f}, {swap_energy:>6.2f}  "
            "(difference: {diff_energy:>6.2f}) at pH {ph:>6.2f}\n"
            " Interaction energy:  {int_energy:>6.2f}\n"
            " Intrinsic pka's:     {pka1:>6.2f}, {pka2:>6.2f}  "
            "(difference: {diff_pka:>6.2f})\n"
            " Swapped pKa's:       {swap1:>6.2f}, {swap2:>6.2f}  "
            "(difference: {shift1:>6.2f}, {shift2:>6.2f})"
        ).format(
            label1=group1.label, label2=group2.label,
            coupl_fact=data['coupling_factor'],
            def_energy=data['default_energy'],
            swap_energy=data['swapped_energy'],
            diff_energy=data['default_energy']-data['swapped_energy'],
            ph=data['pH'], int_energy=data['interaction_energy'],
            pka1=group1.intrinsic_pka, pka2=group2.intrinsic_pka,
            diff_pka=group1.intrinsic_pka-group2.intrinsic_pka,
            swap1=data['swapped_pka1'], swap2=data['swapped_pka2'],
            shift1=data['pka_shift1'], shift2=data['pka_shift2'])
        return str_


NCCG = NonCovalentlyCoupledGroups()



"""Implements many of the main functions used to call PROPKA."""
import sys
import logging
import argparse
import pkg_resources


_LOGGER = logging.getLogger("propka")
_STDOUT_HANDLER = logging.StreamHandler(sys.stdout)
_STDOUT_HANDLER.setFormatter(logging.Formatter("%(message)s"))
_LOGGER.addHandler(_STDOUT_HANDLER)


EXPECTED_ATOM_NUMBERS = {'ALA': 5, 'ARG': 11, 'ASN': 8, 'ASP': 8, 'CYS': 6,
                         'GLY': 4, 'GLN': 9, 'GLU': 9, 'HIS': 10, 'ILE': 8,
                         'LEU': 8, 'LYS': 9, 'MET': 8, 'PHE': 11, 'PRO': 7,
                         'SER': 6, 'THR': 7, 'TRP': 14, 'TYR': 12, 'VAL': 7}


def protein_precheck(conformations, names):
    """Check protein for correct number of atoms, etc.

    Args:
        names:  conformation names to check
    """
    for name in names:
        atoms = conformations[name].atoms
        # Group the atoms by their residue:
        atoms_by_residue = {}
        for atom in atoms:
            if atom.element != 'H':
                res_id = resid_from_atom(atom)
                try:
                    atoms_by_residue[res_id].append(atom)
                except KeyError:
                    atoms_by_residue[res_id] = [atom]
        for res_id, res_atoms in atoms_by_residue.items():
            res_name = res_atoms[0].res_name
            residue_label = '{0:>3s}{1:>5s}'.format(res_name, res_id)
            # ignore ligand residues
            if res_name not in EXPECTED_ATOM_NUMBERS:
                continue
            # check for c-terminal
            if 'C-' in [a.terminal for a in res_atoms]:
                if len(res_atoms) != EXPECTED_ATOM_NUMBERS[res_name]+1:
                    str_ = ("Unexpected number ({num:d}) of atoms in residue "
                            "{res:s} in conformation {conf:s}".format(
                                num=len(res_atoms), res=residue_label,
                                conf=name))
                    warning(str_)
                continue
            # check number of atoms in residue
            if len(res_atoms) != EXPECTED_ATOM_NUMBERS[res_name]:
                str_ = ("Unexpected number ({num:d}) of atoms in residue "
                        "{res:s} in conformation {conf:s}".format(
                            num=len(res_atoms), res=residue_label,
                            conf=name))
                warning(str_)


def resid_from_atom(atom):
    """Return string with atom residue information.

    Args:
        atom:  atom to generate string for
    Returns
        string
    """
    return '{0:>4d} {1:s} {2:s}'.format(
        atom.res_num, atom.chain_id, atom.icode)


def split_atoms_into_molecules(atoms):
    """Maps atoms into molecules.

    Args:
        atoms:  list of atoms
    Returns:
        list of molecules
    """
    molecules = []
    while len(atoms) > 0:
        initial_atom = atoms.pop()
        molecules.append(make_molecule(initial_atom, atoms))
    return molecules


def make_molecule(atom, atoms):
    """Make a molecule from atoms.

    Args:
        atom:  one of the atoms
        atoms:  a list of the remaining atoms
    Return:
        list of atoms
    """
    bonded_atoms = [a for a in atoms if atom in a.bonded_atoms]
    res_atoms = [atom,]
    for bond_atom in bonded_atoms:
        if bond_atom in atoms:
            atoms.remove(bond_atom)
            res_atoms.extend(make_molecule(bond_atom, atoms))
    return res_atoms


def make_grid(min_, max_, step):
    """Make a grid across the specified tange.

    TODO - figure out if this duplicates existing generators like `range` or
    numpy function.

    Args:
        min_:  minimum value of grid
        max_:  maximum value of grid
        step:  grid step size
    """
    x = min_
    while x <= max_:
        yield x
        x += step


def generate_combinations(interactions):
    """Generate combinations of interactions.

    Args:
        interactions:  list of interactions
    Returns:
        list of combinations
    """
    res = [[]]
    for interaction in interactions:
        res = make_combination(res, interaction)
    res.remove([])
    return res


def make_combination(combis, interaction):
    """Make a specific set of combinations.

    Args:
        combis:  list of combinations
        interaction:  interaction to add to combinations
    Returns:
        list of combinations
    """
    res = []
    for combi in combis:
        res.append(combi+[interaction])
        res.append(combi)
    return res


def parse_res_string(res_str):
    """Parse a residue string.

    Args:
        res_string:  residue string in format "chain:resnum[inscode]"
    Returns:
        a tuple of (chain, resnum, inscode).
    Raises:
        ValueError if the input string is invalid.
    """
    try:
        chain, resnum_str = res_str.split(":")
    except ValueError:
        raise ValueError("Invalid residue string (must contain 2 "
                         "colon-separated values)")
    try:
        resnum = int(resnum_str)
    except ValueError:
        try:
            resnum = int(resnum_str[:-1])
        except ValueError:
            raise ValueError("Invalid residue number (not an int)")
        else:
            inscode = resnum_str[-1]
    else:
        inscode = " "
    return chain, resnum, inscode


def build_parser(parser=None):
    """Build an argument parser for PROPKA.

    Args:
        parser:  existing parser. If this is not None, then the PROPKA parser will
                 be created as a subparser to this existing parser.  Otherwise, a
                 new parser will be created.
    Returns:
        ArgumentParser object.
    """
    if parser is not None:
        group = parser.add_argument_group(title="PROPKA invoation options")
    else:
        parser = argparse.ArgumentParser(
            description=("PROPKA predicts the pKa values of ionizable "
                         "groups in proteins and protein-ligand "
                         "complexes based in the 3D structure"),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        # This is duck-typing at its finest
        group = parser
        group.add_argument("input_pdb", help="read data from <filename>")
    group.add_argument(
        "-f", "--file", action="append", dest="filenames", default=[],
        help="read data from <filename>, i.e. <filename> is added to arguments")
    group.add_argument(
        "-r", "--reference", dest="reference", default="neutral",
        help=("setting which reference to use for stability calculations "
              "[neutral/low-pH]"))
    group.add_argument(
        "-c", "--chain", action="append", dest="chains",
        help=('creating the protein with only a specified chain. Specify '
              '" " for chains without ID [all]'))
    group.add_argument(
        "-i", "--titrate_only", dest="titrate_only",
        help=('Treat only the specified residues as titratable. Value should '
              'be a comma-separated list of "chain:resnum" values; for example: '
              '-i "A:10,A:11"'))
    group.add_argument(
        "-t", "--thermophile", action="append", dest="thermophiles",
        help=("defining a thermophile filename; usually used in "
              "'alignment-mutations'"))
    group.add_argument(
        "-a", "--alignment", action="append", dest="alignment",
        help=("alignment file connecting <filename> and <thermophile> "
              "[<thermophile>.pir]"))
    group.add_argument(
        "-m", "--mutation", action="append", dest="mutations",
        help=("specifying mutation labels which is used to modify "
              "<filename> according to, e.g. N25R/N181D"))
    group.add_argument(
        "-v", "--version", dest="version_label", default="Jan15",
        help="specifying the sub-version of propka [Jan15/Dec19]")
    group.add_argument(
        "-p", "--parameters", dest="parameters",
        default=pkg_resources.resource_filename(__name__, "propka.cfg"),
        help="set the parameter file [{default:s}]")
    try:
        group.add_argument(
            "--log-level",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="logging level verbosity", default="INFO")
    except argparse.ArgumentError:
        # It is possible that --log-level has already been set by APBS
        pass
    group.add_argument(
        "-o", "--pH", dest="pH", type=float, default=7.0,
        help="setting pH-value used in e.g. stability calculations [7.0]")
    group.add_argument(
        "-w", "--window", dest="window", nargs=3, type=float,
        default=(0.0, 14.0, 1.0),
        help=("setting the pH-window to show e.g. stability profiles "
              "[0.0, 14.0, 1.0]"))
    group.add_argument(
        "-g", "--grid", dest="grid", nargs=3, type=float,
        default=(0.0, 14.0, 0.1),
        help=("setting the pH-grid to calculate e.g. stability "
              "related properties [0.0, 14.0, 0.1]"))
    group.add_argument(
        "--mutator", dest="mutator",
        help="setting approach for mutating <filename> [alignment/scwrl/jackal]")
    group.add_argument(
        "--mutator-option", dest="mutator_options", action="append",
        help="setting property for mutator [e.g. type=\"side-chain\"]")
    group.add_argument(
        "-d", "--display-coupled-residues", dest="display_coupled_residues",
        action="store_true",
        help=("Displays alternative pKa values due "
              "to coupling of titratable groups"))
    group.add_argument(
        "-l", "--reuse-ligand-mol2-files", dest="reuse_ligand_mol2_file",
        action="store_true", default=False,
        help=("Reuses the ligand mol2 files allowing the user to alter "
              "ligand bond orders"))
    group.add_argument(
        "-k", "--keep-protons", dest="keep_protons", action="store_true",
        help="Keep protons in input file", default=False)
    group.add_argument(
        "-q", "--quiet", action="store_const", const="WARNING",
        dest="log_level", help="suppress non-warning messages")
    group.add_argument(
        "--generate-propka-input", action="store_true",
        help="Generate a PROPKA input file")
    group.add_argument(
        "--protonate-all", dest="protonate_all", action="store_true",
        help="Protonate all atoms (will not influence pKa calculation)",
        default=False)
    return parser


def loadOptions(args=None):
    """
    Load the arguments parser with options. Note that verbosity is set as soon
    as this function is invoked.

    Arguments:
        args:  list of arguments
    Returns:
        argparse namespace
    """
    # loading the parser
    parser = build_parser()
    # parsing and returning options and arguments
    options = parser.parse_args(args)

    # adding specified filenames to arguments
    options.filenames.append(options.input_pdb)
    # Convert titrate_only string to a list of (chain, resnum) items:
    if options.titrate_only is not None:
        res_list = []
        for res_str in options.titrate_only.split(','):
            try:
                chain, resnum, inscode = parse_res_string(res_str)
            except ValueError:
                _LOGGER.critical(
                    'Invalid residue string: "{0:s}"'.format(res_str))
                sys.exit(1)
            res_list.append((chain, resnum, inscode))
        options.titrate_only = res_list
    # Set the no-print variable
    level = getattr(logging, options.log_level)
    _LOGGER.setLevel(level)
    # done!
    return options


def make_tidy_atom_label(name, element):
    """Returns a 'tidier' atom label for printing to the new PDB file.

    Args:
        name:  atom name
        element:  atom element
    Returns:
        string
    """
    if len(name) > 4: # if longer than 4, just truncate the name
        label = name[0:4]
    elif len(name) == 4: # if length is 4, otherwise use the name as it is
        label = name
    else: # if less than 4 characters long, insert white space as needed
        if len(element) == 1:
            label = ' {0:<3s}'.format(name)
        else: # The element should occupy the two first chars
            label = '{0:<4s}'.format(name)
    return label


def get_sorted_configurations(configuration_keys):
    """Extract and sort configurations.

    Args:
        configuration_keys:  list of configuration keys
    Returns:
        list of configurations
    """
    configurations = list(configuration_keys)
    configurations.sort(key=configuration_compare)
    return configurations


def configuration_compare(conf):
    """TODO - figure out what this function does."""
    return 100*int(conf[1:-2]) + ord(conf[-1])


def _args_to_str(arg_list):
    """Summarize list of arguments in string.

    Args:
        arg_list:  list of arguments
    Returns:
        string
    """
    return " ".join(map(str, arg_list))


def info(*args):
    """Log a message to info.

    Level defaults to INFO unless overridden.

    Args:
        args:  argument list
    """
    _LOGGER.info(_args_to_str(args))


def debug(*args):
    """Log a message to debug.

    Level defaults to DEBUG unless overridden.

    Args:
        args:  argument list
    """
    _LOGGER.debug(_args_to_str(args))


def warning(*args):
    """Log a message to warning.

    Level defaults to WARNING unless overridden.

    Args:
        args:  argument list
    """
    _LOGGER.warning(_args_to_str(args))



#! /usr/bin/python
# PROPKA 3

from setuptools import setup, find_packages

VERSION = "3.2.0"

setup(
    name="propka",
    version=VERSION,
    description="Heuristic pKa calculations with ligands", long_description="""
PROPKA predicts the pKa values of ionizable groups in proteins (version 3.0) and
protein-ligand complexes (version 3.1 and later) based on the 3D structure.

For proteins without ligands both version should produce the same result.

The method is described in the following papers, which you should cite
in publications:

* Sondergaard, Chresten R., Mats HM Olsson, Michal Rostkowski, and Jan
  H. Jensen. "Improved Treatment of Ligands and Coupling Effects in
  Empirical Calculation and Rationalization of pKa Values." Journal of
  Chemical Theory and Computation 7, no. 7 (2011): 2284-2295.

* Olsson, Mats HM, Chresten R. Sondergaard, Michal Rostkowski, and Jan
  H. Jensen. "PROPKA3: consistent treatment of internal and surface
  residues in empirical pKa predictions." Journal of Chemical Theory
  and Computation 7, no. 2 (2011): 525-537.

See http://propka.org/ for the PROPKA web server.
""",
    author="Jan H. Jensen",
    author_email="jhjensen@chem.ku.dk",
    maintainer="Nathan Baker",
    maintainer_email="nathanandrewbaker@gmail.com",
    license="LGPL v2.1",
    url="http://propka.org",
    keywords="science",
    classifiers=[
        'Development Status :: 6 - Mature',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Operating System :: POSIX', 
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        ],
    packages=find_packages(exclude=['scripts']),
    package_data={'propka': ['*.dat', '*.cfg', '*.json']},
    entry_points={'console_scripts': ['propka3 = propka.run:main', ]},
    zip_safe=True,
    python_requires='>=3.6',
    tests_require=["pandas", "numpy", "pytest"],
    test_suite="tests",
    )



"""Entry point for PROPKA script."""
import logging
from propka.lib import loadOptions
from propka.input import read_parameter_file, read_molecule_file
from propka.parameters import Parameters
from propka.molecular_container import MolecularContainer


_LOGGER = logging.getLogger("PROPKA")


def main(optargs=None):
    """Read in structure files, calculate pKa values, and print pKa files."""
    # loading options, flags and arguments
    optargs = optargs if optargs is not None else []
    options = loadOptions(*optargs)
    pdbfiles = options.filenames
    parameters = read_parameter_file(options.parameters, Parameters())
    for pdbfile in pdbfiles:
        my_molecule = MolecularContainer(parameters, options)
        my_molecule = read_molecule_file(pdbfile, my_molecule)
        my_molecule.calculate_pka()
        my_molecule.write_pka()
        if options.generate_propka_input:
            my_molecule.write_propka()


def single(pdbfile, optargs=None):
    """Run a single PROPKA calculation using *pdbfile* as input.

    Commandline options can be passed as a **list** in *optargs*.

    .. rubric:: Example

    ::
       single("protein.pdb", optargs=["--mutation=N25R/N181D", "-v",
              "--pH=7.2"])
    """
    optargs = optargs if optargs is not None else []
    options = loadOptions(*optargs)
    pdbfile = options.filenames.pop(0)
    parameters = read_parameter_file(options.parameters, Parameters())
    if len(options.filenames) > 0:
        _LOGGER.warning("Ignoring filenames: {0:s}".format(options.filenames))
    my_molecule = MolecularContainer(parameters, options)
    my_molecule = read_molecule_file(pdbfile, my_molecule)
    my_molecule.calculate_pka()
    my_molecule.write_pka()
    if options.generate_propka_input:
        my_molecule.write_propka()
    return my_molecule



"""Contains version-specific methods and parameters.

TODO - this module unnecessarily confuses the code.  Can we eliminate it?
"""
from propka.lib import info
from propka.hydrogens import setup_bonding_and_protonation, setup_bonding
from propka.hydrogens import setup_bonding_and_protonation_30_style
from propka.energy import radial_volume_desolvation, calculate_pair_weight
from propka.energy import hydrogen_bond_energy, hydrogen_bond_interaction
from propka.energy import electrostatic_interaction, check_coulomb_pair
from propka.energy import coulomb_energy, check_exceptions
from propka.energy import backbone_reorganization


class Version:
    """Store version-specific methods and parameters."""
    def __init__(self, parameters):
        self.parameters = parameters
        self.desolvation_model = self.empty_function
        self.weight_pair_method = self.empty_function
        self.hydrogen_bond_interaction_model = self.empty_function
        self.sidechain_interaction_model = self.empty_function
        self.electrostatic_interaction_model = self.empty_function
        self.coulomb_interaction_model = self.empty_function
        self.check_coulomb_pair_method = self.empty_function
        self.backbone_reorganisation_method = self.empty_function
        self.exception_check_method = self.empty_function
        self.molecular_preparation_method = self.empty_function
        self.prepare_bonds = self.empty_function

    @staticmethod
    def empty_function(*args):
        """Placeholder function so we don't use uninitialized variables.

        Args:
            args:  whatever arguments would have been passed to the function
        Raises:
            NotImplementedError
        """
        err = "Called an empty Version function with args {0:s}".format(args)
        raise NotImplementedError(err)

    def calculate_desolvation(self, group):
        """Calculate desolvation energy using assigned model."""
        return self.desolvation_model(self.parameters, group)

    def calculate_pair_weight(self, num_volume1, num_volume2):
        """Calculate pair weight using assigned model."""
        return self.weight_pair_method(
            self.parameters, num_volume1, num_volume2)

    def hydrogen_bond_interaction(self, group1, group2):
        """Calculate H-bond energy using assigned model."""
        return self.hydrogen_bond_interaction_model(group1, group2, self)

    def calculate_side_chain_energy(self, distance, dpka_max, cutoff, _,
                                    f_angle):
        """Calculate sidechain energy using assigned model."""
        return self.sidechain_interaction_model(
            distance, dpka_max, cutoff, f_angle)

    def electrostatic_interaction(self, group1, group2, distance):
        """Calculate electrostatic energy using assigned model."""
        return self.electrostatic_interaction_model(
            group1, group2, distance, self)

    def calculate_coulomb_energy(self, distance, weight):
        """Calculate Coulomb energy using assigned model."""
        return self.coulomb_interaction_model(
            distance, weight, self.parameters)

    def check_coulomb_pair(self, group1, group2, distance):
        """Check Coulomb pair using assigned model."""
        return self.check_coulomb_pair_method(
            self.parameters, group1, group2, distance)

    def calculate_backbone_reorganization(self, conformation):
        """Calculate backbone reorganization using assigned model."""
        return self.backbone_reorganisation_method(
            self.parameters, conformation)

    def check_exceptions(self, group1, group2):
        """Calculate exceptions using assigned model."""
        return self.exception_check_method(self, group1, group2)

    def setup_bonding_and_protonation(self, molecular_container):
        """Setup bonding and protonation using assigned model."""
        return self.molecular_preparation_method(molecular_container)

    def setup_bonding(self, molecular_container):
        """Setup bonding using assigned model."""
        return self.prepare_bonds(self.parameters, molecular_container)


class VersionA(Version):
    """TODO - figure out what this is."""

    def __init__(self, parameters):
        """Initialize object with parameters."""
        # set the calculation rutines used in this version
        super().__init__(parameters)
        self.molecular_preparation_method = setup_bonding_and_protonation
        self.prepare_bonds = setup_bonding
        self.desolvation_model = radial_volume_desolvation
        self.weight_pair_method = calculate_pair_weight
        self.sidechain_interaction_model = hydrogen_bond_energy
        self.hydrogen_bond_interaction_model = hydrogen_bond_interaction
        self.electrostatic_interaction_model = electrostatic_interaction
        self.check_coulomb_pair_method = check_coulomb_pair
        self.coulomb_interaction_model = coulomb_energy
        self.backbone_interaction_model = hydrogen_bond_energy
        self.backbone_reorganisation_method = backbone_reorganization
        self.exception_check_method = check_exceptions

    def get_hydrogen_bond_parameters(self, atom1, atom2):
        """Get hydrogen bond parameters for two atoms.

        Args:
            atom1:  first atom
            atom2:  second atom
        Returns:
            [dpka_max, cutoff]
        """
        dpka_max = self.parameters.sidechain_interaction
        cutoff = self.parameters.sidechain_cutoffs.get_value(
            atom1.group_type, atom2.group_type)
        return [dpka_max, cutoff]

    def get_backbone_hydrogen_bond_parameters(self, backbone_atom, atom):
        """Get hydrogen bond parameters between backbone atom and other atom.

        Args:
            backbone_atom:  backbone atom
            atom:  other atom
        Returns
            [v, [c1, c3]]  TODO - figure out what this is
        """
        if backbone_atom.group_type == 'BBC':
            if atom.group_type in self.parameters.backbone_CO_hydrogen_bond.keys():
                [v, c1, c2] = self.parameters.backbone_CO_hydrogen_bond[
                    atom.group_type]
                return [v, [c1, c2]]
        if backbone_atom.group_type == 'BBN':
            if atom.group_type in self.parameters.backbone_NH_hydrogen_bond.keys():
                [v, c1, c2] = self.parameters.backbone_NH_hydrogen_bond[
                    atom.group_type]
                return [v, [c1, c2]]
        return None


class SimpleHB(VersionA):
    """A simple hydrogen bond version."""

    def __init__(self, parameters):
        """Initialize object with parameters."""
        # set the calculation rutines used in this version
        super().__init__(parameters)
        info('Using simple hb model')

    def get_hydrogen_bond_parameters(self, atom1, atom2):
        """Get hydrogen bond parameters for two atoms.

        Args:
            atom1:  first atom
            atom2:  second atom
        Returns:
            [dpka_max, cutoff]
        """
        return self.parameters.hydrogen_bonds.get_value(
            atom1.element, atom2.element)

    def get_backbone_hydrogen_bond_parameters(self, backbone_atom, atom):
        """Get hydrogen bond parameters between backbone atom and other atom.

        Args:
            backbone_atom:  backbone atom
            atom:  other atom
        Returns
            [v, [c1, c3]]  TODO - figure out what this is
        """
        return self.parameters.hydrogen_bonds.get_value(
            backbone_atom.element, atom.element)


class ElementBasedLigandInteractions(VersionA):
    """TODO - figure out what this is."""

    def __init__(self, parameters):
        """Initialize object with parameters."""
        # set the calculation rutines used in this version
        super().__init__(parameters)
        info('Using detailed SC model!')
        return

    def get_hydrogen_bond_parameters(self, atom1, atom2):
        """Get hydrogen bond parameters for two atoms.

        Args:
            atom1:  first atom
            atom2:  second atom
        Returns:
            [dpka_max, cutoff]
        """
        if 'hetatm' not in [atom1.type, atom2.type]:
            # this is a protein-protein interaction
            dpka_max = self.parameters.sidechain_interaction.get_value(
                atom1.group_type, atom2.group_type)
            cutoff = self.parameters.sidechain_cutoffs.get_value(
                atom1.group_type, atom2.group_type)
            return [dpka_max, cutoff]
        # at least one ligand atom is involved in this interaction
        # make sure that we are using the heavy atoms for finding paramters
        elements = []
        for atom in [atom1, atom2]:
            if atom.element == 'H':
                elements.append(atom.bonded_atoms[0].element)
            else:
                elements.append(atom.element)
        return self.parameters.hydrogen_bonds.get_value(
            elements[0], elements[1])

    def get_backbone_hydrogen_bond_parameters(self, backbone_atom, atom):
        """Get hydrogen bond parameters between backbone atom and other atom.

        Args:
            backbone_atom:  backbone atom
            atom:  other atom
        Returns
            [v, [c1, c3]]  TODO - figure out what this is
        """
        if atom.type == 'atom':
            # this is a backbone-protein interaction
            if (backbone_atom.group_type == 'BBC'
                    and atom.group_type
                    in self.parameters.backbone_CO_hydrogen_bond.keys()):
                [v, c1, c2] = self.parameters.backbone_CO_hydrogen_bond[
                    atom.group_type]
                return [v, [c1, c2]]

            if (backbone_atom.group_type == 'BBN'
                    and atom.group_type
                    in self.parameters.backbone_NH_hydrogen_bond.keys()):
                [v, c1, c2] = self.parameters.backbone_NH_hydrogen_bond[
                    atom.group_type]
                return [v, [c1, c2]]
        else:
            # this is a backbone-ligand interaction
            # make sure that we are using the heavy atoms for finding paramters
            elements = []
            for atom2 in [backbone_atom, atom]:
                if atom2.element == 'H':
                    elements.append(atom2.bonded_atoms[0].element)
                else:
                    elements.append(atom2.element)
            res = self.parameters.hydrogen_bonds.get_value(
                elements[0], elements[1])
            if not res:
                info(
                    'Could not determine backbone interaction parameters for:',
                    backbone_atom, atom)
            return None
        return None


class Propka30(Version):
    """Version class for PROPKA 3.0."""

    def __init__(self, parameters):
        """Initialize object with parameters."""
        # set the calculation routines used in this version
        super().__init__(parameters)
        self.molecular_preparation_method = (
            setup_bonding_and_protonation_30_style)
        self.desolvation_model = radial_volume_desolvation
        self.weight_pair_method = calculate_pair_weight
        self.sidechain_interaction_model = hydrogen_bond_energy
        self.check_coulomb_pair_method = check_coulomb_pair
        self.coulomb_interaction_model = coulomb_energy
        self.backbone_reorganisation_method = backbone_reorganization
        self.exception_check_method = check_exceptions

    def get_hydrogen_bond_parameters(self, atom1, atom2):
        """Get hydrogen bond parameters for two atoms.

        Args:
            atom1:  first atom
            atom2:  second atom
        Returns:
            [dpka_max, cutoff]
        """
        dpka_max = self.parameters.sidechain_interaction.get_value(
            atom1.group_type, atom2.group_type)
        cutoff = self.parameters.sidechain_cutoffs.get_value(
            atom1.group_type, atom2.group_type)
        return [dpka_max, cutoff]
